#!/usr/bin/perl -w

=pod
 D. Brooks 6/2001

 emapi.pl for estimating ancestral amino acid composition at root of tree using
 EM
 
 usage: 
 
   emapi.pl <alignment> <treefile> 
 
 options:
 
  --threshold <x>   : threshold for convergence 
  		       (default = 0.0001)
  --start_pi <file> : file to read in starting amino acid composition 
                      (default start_pi)
  --end_pi <file>   : file to output estimated ancestral amino acid composition 
                      (default end_pi)
		      

alignment must be in phylip format with sequence ids sequential #'s starting at 1
tree must be in Newick format with branch lengths in integers

note that must have present in the same directory
  jones/*.mat where *.mat are jtt conditional matrices from 0 to 300
 
 order of aa for amino acid composition and matrix input:
 A  R  N  D  C  Q  E  G  H  I  L  K  M  F  P  S  T  W  Y  V
 
 note that most of the computation is carried out in log space
 
=cut 

use Getopt::Long;
use IO::Handle;

GetOptions( "threshold=s" => \$threshold,
            "end_pi=s" => \$end_pi,
            "start_pi=s"  => \$start_pi );

unless ($threshold) {$threshold = 0.0001;} 
unless ($start_pi) {$start_pi = "start_pi"};
unless ($end_pi) {$end_pi = "end_pi"};
 
@aa = qw(A  R  N  D  C  Q  E  G  H  I  L  K  M  F  P  S  T  W  Y  V);
@aatoindex{@aa} = (0..19);
@indextoaa{0..19} = (@aa);

#**********************************************************************
	
#			assign nodes and branch lengths

#**********************************************************************

$align_file = $ARGV[0];
$tree_file = $ARGV[1];

print "\nprocessing tree...\n\n";

open (TREE, "$tree_file")|| die "Sorry couldn't open $tree_file\n";

while (<TREE>) {
	chomp;
	
	unless (/\w/) { next;} # ignores empty lines
	$left = tr/\(//;
	$right = tr/\)//;
	if ($left != $right)  {	
	die "tree format error: unequal ()\n";	
	}
	$intnodes = tr/\,//;
	$nseqs = $intnodes + 1;
	$innernode = $nseqs + 1;

  	# using post-order tree traversal 
	# substitute parent node # for (leftleaf:distance, rightleaf: distance)	
		
 	while (s/\(\s*(\d{1,2})\s*:\s*0*(\.*\d{1,})\s*\,\s*(\d{1,2})\s*:\s*0*(\.*\d{1,})\s*\)/$innernode/) {
	
	$length{$1} = $2;
	$parent{$1} = $innernode;
	$length{$3} = $4;
	$parent{$3} = $innernode;
	@children= ($1, $3);
	push @nodes, @children; # keep track of all nodes
	$children{$innernode} = [@children];
	$innernode++;
	}
}
	
$lcanode = $intnodes + $nseqs;

push @nodes, $lcanode; # adds root to @nodes

close (TREE);

open (TREE, "$tree_file");

@species = ();
while (<TREE>) {
	while (s/\(([a-zA-Z0-9]*)://) {
		push @species, $1;
	}
	while (s/\,([a-zA-Z0-9]*)://) {
		push @species, $1;
	}
}
close (TREE);

#**********************************************************************
	
# 			read in alignment

#**********************************************************************

print "reading alignment...\n\n";
open(FILE, "$align_file") || die "couldn't open $align_file\n";

$count = 0;
$nspecies = @species;

foreach $node (@nodes) {
	$seq{$node} = "";
}

$line = 0;
 
while (<FILE>) {
   chomp;

    unless (/^[A-Za-z]/ || /\(/) {
   		if (/^\S/) {
   		    foreach $species (@species) {
       
    		   if (/^$species /) {
   		 		   s/$species//;
    			   s/ //g;
    			   $seq{$species} .= $_;
    			   $count{$line} = $species;
    			   last;
    	   		}
       		}
   		}      	   
   		if ($line == $nspecies) {
    		   $line = 0;
   		}
   		else {$line++;}
   	}
}

close (FILE);
foreach $species(@species) {
	$seq{$species} =~ s/\s//g;
	$seq{$species} =~ tr/a-z/A-Z/;
}

$seqlength = length($seq{$nodes[0]});

#**********************************************************************
	
# 	   read in intial p(i) and convert to log space

#**********************************************************************

open (FILE, "$start_pi") || die "Sorry couldn't open $start_pi\n";

while (<FILE>) {
	@p = split;
}

foreach $aa (@aa) 
{
	$lnp[$aatoindex{$aa}] = log ($p[$aatoindex{$aa}]);
}

#**********************************************************************
	
# 	determine p(x|i) for each position in sequence


#**********************************************************************
print "determining p(x|i) for each alignment column...\n\n";

for ($pos = 0; $pos < $seqlength; $pos++) {
	if ($pos % 50 == 0) { print "$pos\n";}
	for ($node = 0; $node < @nodes; $node++) {  # assign probs to all nodes

		if ($nodes[$node] <= $nseqs) { # special case: leaves of tree

		$res = substr($seq{$nodes[$node]}, $pos, 1);
			
			foreach $aa (@aa) {
				$lnpsubtree{$nodes[$node]}{$aa} = -10000.; 
			}
			
			$lnpsubtree{$nodes[$node]}{$res} = 0.0; # special case
		}
	
		elsif ($nodes[$node] == $lcanode) { # special case : root
			foreach $aa (@aa) {
			 	$left = lnsumlft($nodes[$node],$aa);
			 	$right = lnsumrt($nodes[$node], $aa);	
				$lnpxgi{$pos}{$aa} = $left + $right; 				
			}
		}
		else {
			foreach $aa (@aa) {
				$lnpsubtree{$nodes[$node]}{$aa} = lnsumlft($nodes[$node], $aa) + lnsumrt($nodes[$node], $aa);
			}
		}
	}
}

print "\n";

=pod
**********************************************************************
	
    EM:
    
     until successive lnP(x) differ by less than some threshold, 
     alternately  
     1)  calculate P(i,x) and P(x) and P(i|x) in each position for a given P(i)
     2) re-estimate P(i)
     
**********************************************************************
=cut

print "iteratively estimating ancestral p(i)...\n\n";

$logscore[0] = -999999999999;
$logscore[1] = -999999999998;

$iteration = 2;

while ($logscore[$iteration - 1] - $logscore[$iteration - 2] > $threshold)
{

#**********************************************************************
	
#    for a given P(i), calculate P(i,x) and P(x) and P(i|x) in each position

#**********************************************************************
        $logscore[$iteration] = 0;
	for ($pos = 0; $pos < $seqlength; $pos++) 
	{
		$lnpx{$pos} = -10000;
		foreach $aa (@aa)
		{
			$lnpiandx{$pos}{$aa} = $lnp[$aatoindex{$aa}] + $lnpxgi{$pos}{$aa};
			$lnpx{$pos} = LogSum($lnpx{$pos}, $lnpiandx{$pos}{$aa});
		}
		$logscore[$iteration] += $lnpx{$pos};
		foreach $aa (@aa)
		{
			$lnpigx{$pos}{$aa} = $lnpiandx{$pos}{$aa} - $lnpx{$pos};
		}	
	}
	# output to screen for happiness
	$n = $iteration - 1;
	print "$n logscore = $logscore[$iteration] \n";


#**********************************************************************
	
# 		reestimate P(i) from sequence at root node

#**********************************************************************

	foreach $aa (@aa){	$sump{$aa} = 0; }

	for ($pos = 0; $pos < $seqlength; $pos++) 
	{
		foreach $aa (@aa)
			{
				$sump{$aa} += exp ($lnpigx{$pos}{$aa});
			}
	}

	foreach $aa (@aa) {
		$p[$aatoindex{$aa}] = $sump{$aa}/ $seqlength;
		if ($p[$aatoindex{$aa}] < 0.0001 ) {$p[$aatoindex{$aa}] = 0.0001 }
		$lnp[$aatoindex{$aa}] = log($p[$aatoindex{$aa}]);
	}
	$iteration++;
}
$iteration -= 2;

print "\nstarting logscore = $logscore[2]\n";
print "ending logscore = $logscore[$iteration]\n";
print "$iteration iterations\n\n";

#**********************************************************************
	
# 		print output

#**********************************************************************

open (ENDPI, ">$end_pi") || die "Sorry couldn't open $end_pi$!\n";

foreach $aa (@aa) {
	printf ENDPI "$aa %4.4f \n", $p[$aatoindex{$aa}];
}
print ENDPI "\n";

#**********************************************************************
	
# 				subroutines

#**********************************************************************

sub lnsumlft {
	my ($i, $j, $lnsum, $node, $aa);

	@pchange = ();	
	
	$lnsum = -10000.0;
	
	$aa = $_[1];
	$i = $aatoindex{$aa}; # trans aa to index for matrix
	$node = $_[0];
	open (MATRIX, "jones/mp_$length{$children{$node}[0]}.mat") || die "can't open matrix$!";
	
	while (<MATRIX>) { #make AoA $pchange[j][i]
		push @pchange, [split];
	}
	
	for ($j = 0; $j < 20; $j++) {
            $lnpchange[$j][$i] = log $pchange[$j][$i];
	    $lnsum = LogSum($lnsum, $lnpchange[$j][$i] + $lnpsubtree{$children{$node}[0]}{$indextoaa{$j}});
	}
	return $lnsum;
}

sub lnsumrt {
	my ($i, $j, $lnsum, $node, $aa);

	@pchange = ();
        $lnsum = -10000.0;
       
	$aa = $_[1];
	$i = $aatoindex{$aa}; # trans aa to index for matrix
	$node = $_[0];
	
	open (MATRIX, "jones/mp_$length{$children{$node}[1]}.mat") || die "can't open matrix$!";
	
	while (<MATRIX>) { #make AoA $pchange[j][i]
		push @pchange, [split];
	}

	for ($j = 0; $j < 20; $j++) {
		 $lnpchange[$j][$i] = log $pchange[$j][$i];
		 $lnsum = LogSum ($lnsum, $lnpchange[$j][$i] + $lnpsubtree{$children{$node}[1]}{$indextoaa{$j}});
	}
	return $lnsum;
}

# Function: LogSum()
# borrowed from Sean Eddy's mathsupport.c for HMMER 
# Purpose:  Returns the log of the sum of two log probabilities.
#
#   log(exp(p1)+exp(p2)) = p1 + log(1 + exp(p2-p1)) for p1 > p2
#   Note that this is in natural log space, not log_2.
   
sub LogSum
{
  my ($p1, $p2);
  $p1 = $_[0];
  $p2 = $_[1];
  if ($p1 > $p2)
  {
    if ($p1-$p2 < 50.) 
    {
       return $p1 + log(1. + exp($p2-$p1))
    }
    else
    {
       return $p1;
    }
         }
  else
  {
    if ($p2-$p1 < 50.) 
    {
	return $p2 + log(1. + exp($p1-$p2));
    }
    else
    {
      return $p2;
     }
  }
}	  



